import asyncio
from contextlib import asynccontextmanager
from typing import Type, Dict, Any, Optional

from alembic.config import Config
from sqlalchemy import event
from sqlalchemy.exc import DatabaseError
from sqlalchemy.ext.asyncio import create_async_engine, AsyncConnection, AsyncSession, AsyncEngine
from sqlalchemy.orm import sessionmaker

from tarka.asqla.alembic import get_alembic_config, AlembicHelper
from tarka.asqla.tx import (
    TransactionExecutor,
    RetryableTransactionFactory,
    SessionTransactionExecutor,
    RetryableSessionTransactionFactory,
)


class Database(object):
    engine: AsyncEngine = None
    serializable_engine: AsyncEngine = None
    session_maker: Type[AsyncSession] = None

    def __init__(
        self,
        alembic_dir: str,
        connect_url: str,
        engine_kwargs: Dict[str, Any] = None,  # like echo, connect_args, etc.
        session_maker_kwargs: Dict[str, Any] = None,  # like expire_on_commit, autoflush, etc.
        aiosqlite_serializable_begin: str = "BEGIN",  # serializable tx support workaround for aiosqlite
        tx_retry_delay: Optional[float] = None,  # tx retry delay helps to relieve db pressure at conflicts
    ):
        """
        If aiosqlite is used aiosqlite_serializable_begin="BEGIN IMMEDIATE" could be better, because then no
        conflicts can happen, each tx caller will wait in line to acquire the lock and no resources will be
        wasted retrying. However, this does not work well with session-tx, see TODO: tx.py:105.
        """
        self.alembic_dir = alembic_dir
        self._connect_url = connect_url
        self._engine_kwargs = engine_kwargs
        self._session_maker_kwargs = session_maker_kwargs
        self._aiosqlite_serializable_begin = aiosqlite_serializable_begin
        self._tx_retry_delay = tx_retry_delay
        self.alembic_head_at_startup = ""

    def get_alembic_config(self) -> Config:
        return get_alembic_config(self.alembic_dir, str(self.engine.sync_engine.url))

    async def _init_engine(self):
        """
        If further customization is necessary for the engines, this shall be overridden.
        To switch SQLite to WAL mode for example:

            if self.engine.dialect.name == "sqlite":
                @event.listens_for(self.engine.sync_engine, "connect")
                def _setup_sqlite_connection(dbapi_con, con_record):
                    dbapi_con.execute("PRAGMA journal_mode = WAL;")
                    dbapi_con.execute("PRAGMA synchronous = NORMAL;")
        """
        self.engine = create_async_engine(self._connect_url, **(self._engine_kwargs or {}))
        self.serializable_engine = self.engine.execution_options(isolation_level="SERIALIZABLE")
        if self.engine.driver == "aiosqlite" and self._aiosqlite_serializable_begin:
            # aiosqlite (<=0.20) does not honor the isolation_level with appropriate BEGIN commands, in fact it
            # never issues begin commands, transactions always start implicitly for reads.
            # SQLite itself does not support concurrent mutating transactions in any mode. In WAL journal_mode, there
            # can be parallel and concurrent readers, but writing is always with an exclusive lock. In this sense "any"
            # transaction will actually be serializable provided that an explicit BEGIN is emitted. The actual
            # difference is the point and time that parallel connections encounter locking errors.
            # The workaround (if not disabled) is to explicitly execute a BEGIN command at transaction start.
            # The command can be set for the database (DEFERRED, IMMEDIATE or EXCLUSIVE) when relevant.
            @event.listens_for(self.serializable_engine.sync_engine, "begin")
            def do_begin(conn):
                conn.exec_driver_sql(self._aiosqlite_serializable_begin)

        self.session_maker = sessionmaker(self.engine, class_=AsyncSession, **(self._session_maker_kwargs or {}))

    async def startup(self):
        """
        Initialize engine first, then bootstrap or migrate schema of the database to be up-to-date.
        """
        await self._init_engine()
        alembic_helper = AlembicHelper(self.get_alembic_config())
        retry = 0
        while True:  # handle conflicting migration attempt by parallel workers
            try:
                async with self.engine.begin() as connection:
                    connection: AsyncConnection
                    self.alembic_head_at_startup = await connection.run_sync(alembic_helper.run_strip_output, "current")

                    await self._upgrade(alembic_helper, connection)
                    await self._start_hook(connection)
            except DatabaseError:
                if retry >= 5:
                    raise
                retry += 1
                await asyncio.sleep(0.25)
            else:
                break

    async def _upgrade(self, alembic_helper: AlembicHelper, connection: AsyncConnection):
        if not self.alembic_head_at_startup.endswith(" (head)"):
            await self._pre_upgrade_hook(connection)
            await connection.run_sync(alembic_helper.run, "upgrade", "head")
            await self._post_upgrade_hook(connection)

    async def _pre_upgrade_hook(self, connection: AsyncConnection):
        """
        Place to put custom DDL commands before alembic upgrade.
        Keep in mind that it is usually better to implement all changes as alembic revisions.
        NOTE: This is called before bootstrap (first upgrade) as well.
        """

    async def _post_upgrade_hook(self, connection: AsyncConnection):
        """
        Place to put custom DDL commands after alembic upgrade.
        Keep in mind that it is usually better to implement all changes as alembic revisions.
        """

    async def _start_hook(self, connection: AsyncConnection):
        """
        Place to put custom DDL commands that are executed each time a server starts.
        The connection is currently at the HEAD alembic revision, in the upgrade transaction.
        """

    async def shutdown(self):
        await self.engine.dispose()

    @asynccontextmanager
    async def run(self):
        """
        Convenience wrapper for startup & shutdown.
        """
        await self.startup()
        try:
            yield self
        finally:
            await self.shutdown()

    def serializable_tx(
        self, tx_factory: RetryableTransactionFactory, retry_delay: Optional[float] = None
    ) -> TransactionExecutor:
        """
        Get a transaction executor with SERIALIZABLE isolation_level and automatic retry.
        """
        return TransactionExecutor(
            self.serializable_engine, tx_factory, retry_delay=retry_delay or self._tx_retry_delay
        )

    def session_serializable_tx(
        self, tx_factory: RetryableSessionTransactionFactory, retry_delay: Optional[float] = None
    ) -> SessionTransactionExecutor:
        """
        Get an ORM transaction executor with SERIALIZABLE isolation_level and automatic retry.
        """
        return SessionTransactionExecutor(
            self.session_maker,
            tx_factory,
            engine=self.serializable_engine,
            retry_delay=retry_delay or self._tx_retry_delay,
        )



from __future__ import annotations

import asyncio
import queue
import threading
import time
from contextlib import asynccontextmanager
from functools import partial
from typing import Optional, Sequence, Any, Callable, TypeVar, Type, AsyncContextManager

import wait_for2

from tarka.utility.thread import AbstractThread


def _callback_result(future: asyncio.Future, result):
    if not future.done():
        future.set_result(result)


def _callback_exception(future: asyncio.Future, exc):
    if not future.done():
        future.set_exception(exc)


class AbstractWorkerJob:
    __slots__ = ()

    execute: Callable[[AbstractAioWorker], Any]

    def done(self) -> bool:
        raise NotImplementedError()

    def cancel(self):
        raise NotImplementedError()

    def set_result(self, result):
        raise NotImplementedError()

    def set_exception(self, exc):
        raise NotImplementedError()


class AbstractThreadWorkerJob(AbstractWorkerJob):
    __slots__ = ("event", "result", "state")

    STATE_WAIT = 0
    STATE_RESULT = 1
    STATE_EXCEPTION = 2
    STATE_CANCEL = 3

    def __init__(self):
        self.event = threading.Event()
        self.result = None
        self.state = self.STATE_WAIT

    def done(self) -> bool:
        return bool(self.state)

    def cancel(self):
        if self.state == self.STATE_WAIT:
            self.state = self.STATE_CANCEL
            self.event.set()

    def set_result(self, result):
        self.result = result
        self.state = self.STATE_RESULT
        self.event.set()

    def set_exception(self, exc):
        self.result = exc
        self.state = self.STATE_EXCEPTION
        self.event.set()

    def get_result(self, timeout=None):
        if not self.event.wait(timeout):
            self.cancel()
            raise TimeoutError
        if self.state == self.STATE_WAIT:
            raise Exception("Job result is not set")
        elif self.state == self.STATE_CANCEL:
            raise Exception("Job result is cancelled")
        elif self.state == self.STATE_RESULT:
            return self.result
        elif not isinstance(self.result, Exception):
            raise Exception(f"Job result is not an exception: {self.result}")
        raise self.result


class PartialMethodThreadWorkerJob(AbstractThreadWorkerJob):
    __slots__ = ("impl_fn", "args")

    def __init__(self, impl_fn: Callable[[AbstractAioWorker, ...], Any], args):
        AbstractThreadWorkerJob.__init__(self)
        self.impl_fn = impl_fn
        self.args = args

    def execute(self, worker: AbstractAioWorker) -> Any:
        return self.impl_fn(worker, *self.args)


class AbstractAioWorkerJob(AbstractWorkerJob):
    __slots__ = ("future",)

    def __init__(self, future: asyncio.Future):
        self.future = future

    def done(self) -> bool:
        return self.future.done()

    def cancel(self):
        return self.future.cancel()

    def set_result(self, result):
        self.future.get_loop().call_soon_threadsafe(_callback_result, self.future, result)

    def set_exception(self, exc):
        self.future.get_loop().call_soon_threadsafe(_callback_exception, self.future, exc)


class PartialMethodAioWorkerJob(AbstractAioWorkerJob):
    __slots__ = ("impl_fn", "args")

    def __init__(self, future: asyncio.Future, impl_fn: Callable[[AbstractAioWorker, ...], Any], args):
        AbstractAioWorkerJob.__init__(self, future)
        self.impl_fn = impl_fn
        self.args = args

    def execute(self, worker: AbstractAioWorker) -> Any:
        return self.impl_fn(worker, *self.args)


_ClsT = TypeVar("_ClsT")


class PollTimedQueue(queue.Queue):
    def __init__(self, poll_timeout: float, maxsize=0):
        self.poll_timeout = poll_timeout
        self.next_poll = time.perf_counter() + self.poll_timeout
        queue.Queue.__init__(self, maxsize)

    def get(self, block=True, timeout=None):
        try:
            timeout = self.next_poll - time.perf_counter()
            if timeout < 0.0:  #
                raise queue.Empty
            return queue.Queue.get(self, timeout=timeout)
        except queue.Empty:
            self.next_poll = time.perf_counter() + self.poll_timeout
            raise


class AbstractAioWorker(AbstractThread):
    """
    A lightweight asyncio compatible, customizable interface to an arbitrary thread worker.
    """

    __slots__ = ("_loop", "_poll_timeout", "_request_queue", "started", "closed")

    @classmethod
    @asynccontextmanager
    async def create(cls: Type[_ClsT], *args, **kwargs) -> AsyncContextManager[_ClsT]:
        """
        Strict start-stop helper for automatic cleanup.
        """
        self = cls(*args, **kwargs)
        try:
            self.start()
            await self.wait_ready()
            yield self
        finally:
            self.stop()

    def __init__(self, args: Optional[Sequence[Any]] = None, poll_timeout: Optional[float] = None):
        if poll_timeout is not None and poll_timeout <= 0.0:  # pragma: no cover
            raise ValueError("Aio worker 'poll_timeout' must be greater than zero or None.")
        self._loop = asyncio.get_running_loop()
        self._poll_timeout = poll_timeout
        self._request_queue: queue.Queue = None
        AbstractThread.__init__(self, args)
        self.started = asyncio.Event()
        self.closed = asyncio.Event()

    async def wait_ready(self):
        if self.is_alive():  # only wait for start/stop if the worker thread has been started
            started_task = self._loop.create_task(self.started.wait())
            closed_task = self._loop.create_task(self.closed.wait())
            _, pending = await asyncio.wait([started_task, closed_task], return_when=asyncio.FIRST_COMPLETED)
            if pending:  # cleanup tasks to avoid them being logged as abandoned
                for t in pending:
                    t.cancel()
                await asyncio.gather(*pending, return_exceptions=True)
        if self.closed.is_set() or not self.started.is_set():
            raise Exception(f"Aio worker could not start for {self.__class__.__name__}")

    def start(
        self,
        args: Optional[Sequence[Any]] = None,
        callback: Optional[Callable[[], None]] = None,
        daemon: Optional[bool] = None,
        name_prefix: Optional[str] = None,
    ):
        if callback is not None:  # pragma: no cover
            raise ValueError("Aio worker does not support custom callback. Use the 'closed' event.")
        if daemon not in (None, False):  # pragma: no cover
            raise ValueError("Aio worker must not be daemon to ensure cleanup.")
        self._request_queue = queue.Queue() if self._poll_timeout is None else PollTimedQueue(self._poll_timeout)
        AbstractThread.start(self, None, partial(self._loop.call_soon_threadsafe, self.closed.set), False, name_prefix)

    def stop(self, timeout: Optional[float] = 0) -> bool:
        """
        The default timeout is zero, we assume the .closed event will be used to wait for cleanup in asyncio.
        """
        q = self._request_queue  # local reference resolves race-conditions with worker thread
        if q is not None:
            q.put_nowait(None)
        return AbstractThread.stop(self, timeout)

    def _thread(self, *args: Any) -> None:
        """
        General worker and request executor thread.
        """
        try:
            self._thread_init(*args)
            self._loop.call_soon_threadsafe(self.started.set)
            # run job queue
            while True:
                try:
                    job = self._request_queue.get()
                except queue.Empty:
                    self._thread_poll()
                    continue
                if not isinstance(job, AbstractWorkerJob):
                    break
                if job.done():  # Could have been cancelled before we got to it.
                    continue
                try:
                    result = job.execute(self)
                    job.set_result(result)
                except Exception as e:
                    job.set_exception(e)
        finally:
            # prevent more jobs to be queued
            q = self._request_queue
            self._request_queue = None
            # notify dead jobs if any
            try:
                while True:
                    job = q.get_nowait()
                    if isinstance(job, AbstractWorkerJob):
                        job.cancel()
            except queue.Empty:
                pass
            self._thread_cleanup()

    def _thread_init(self, *args: Any):
        """
        Setup facilities and states for work. The args are passed from __init__ or start.
        """
        raise NotImplementedError()

    def _thread_poll(self):
        """
        Called periodically when poll_timeout is configured.
        """
        raise NotImplementedError()

    def _thread_cleanup(self):
        """
        Cleanup everything before thread exit.
        """
        raise NotImplementedError()

    async def _method_job(self, impl_fn: Callable, *args, timeout: Optional[float] = None) -> Any:
        """
        This is designed to be used as

            get_xy = partialmethod(AbstractAioWorker._method_job, _get_xy_impl)

        which means the wrapped function is not yet bound in the expression. The self argument is automatically
        passed by the worker thread to make it work.
        """
        f = self._loop.create_future()
        self._request_queue.put_nowait(PartialMethodAioWorkerJob(f, impl_fn, args))
        return await wait_for2.wait_for(f, timeout)

    def _method_call(self, impl_fn: Callable, *args, timeout: Optional[float] = None) -> Any:
        """
        This is designed to be used as

            get_xy = partialmethod(AbstractAioWorker._method_call, _get_xy_impl)

        which means the wrapped function is not yet bound in the expression. The self argument is automatically
        passed by the worker thread to make it work.
        """
        job = PartialMethodThreadWorkerJob(impl_fn, args)
        self._request_queue.put_nowait(job)
        return job.get_result(timeout)



import os
import sys
import warnings
from importlib import import_module
from importlib.machinery import EXTENSION_SUFFIXES, SOURCE_SUFFIXES, BYTECODE_SUFFIXES
from importlib.util import spec_from_file_location, module_from_spec
from itertools import chain
from types import ModuleType
from typing import Sequence, Optional
from urllib.parse import quote


def forward_import_recursively(package_module_path: str, skip_dir_names: Sequence[str] = ("__pycache__",)):
    """
    This can be utilized to load plugin-like solutions and structures that utilize subclass-hook for example.
    """
    package_init = import_module(package_module_path).__file__
    package_dir_path = os.path.dirname(package_init)
    if os.path.splitext(os.path.basename(package_init))[0] != "__init__" or not os.path.isdir(package_dir_path):
        return  # Imported module is not a traditional python package (directory with __init__ module)
    for root, dirs, files in os.walk(package_dir_path):
        if os.path.basename(root) in skip_dir_names:
            continue
        imp_root = package_module_path
        sub_path = root[len(package_dir_path) + 1 :]
        if sub_path:
            imp_root = ".".join([imp_root, *sub_path.split(os.path.sep)])
        for file in files:
            if os.path.splitext(file)[0] == "__init__":
                continue
            for suffix in chain(SOURCE_SUFFIXES, BYTECODE_SUFFIXES, EXTENSION_SUFFIXES):
                if file.endswith(suffix):
                    import_module(f"{imp_root}.{file[:-len(suffix)]}")
        for dir_ in dirs:
            if dir_ in skip_dir_names:
                continue
            if any(
                os.path.isfile(os.path.join(root, dir_, f"__init__{suffix}"))
                for suffix in chain(SOURCE_SUFFIXES, BYTECODE_SUFFIXES)
            ):
                import_module(f"{imp_root}.{dir_}")


def import_optional(import_path: str, warn_error: bool = True) -> Optional[ModuleType]:
    try:
        return import_module(import_path)
    except ImportError:
        if warn_error:
            warnings.warn(f"Can't import optional {import_path}")


class IsolatedPackageImport:
    """
    This can be used to import modules on an arbitrary absolute path and have them all under an isolated package if
    they are imported with relative paths.
    """

    @classmethod
    def create(cls, root_module_path: str):
        # create fake package (for relative imports)
        package_name = "__main__" + quote(root_module_path, safe="").replace(".", "_")
        package_module = ModuleType(package_name)
        package_module.__path__ = [os.path.dirname(root_module_path)]
        package_module.__file__ = os.path.join(package_module.__path__[0], "__init__.py")
        sys.modules[package_name] = package_module
        try:
            # import main module under fake package
            module_name = package_name + "." + os.path.basename(root_module_path).rpartition(".")[0]
            spec = spec_from_file_location(module_name, root_module_path)
            main_module = module_from_spec(spec)
            sys.modules[module_name] = main_module
            try:
                spec.loader.exec_module(main_module)
            except:
                del sys.modules[module_name]
                raise
        except:
            del sys.modules[package_name]
            raise
        return cls(package_name, main_module)

    def __init__(self, pacakge_name: str, module: ModuleType):
        self.package_name = pacakge_name
        self.module = module

    def unload(self):
        # NOTE: This only ensures that a subsequent import will be clean, provided that all submodules were imported
        # with relative imports.
        for k in list(sys.modules.keys()):
            if k.startswith(self.package_name):
                if len(k) == len(self.package_name) or k[len(self.package_name)] == ".":
                    del sys.modules[k]



from importlib.machinery import SOURCE_SUFFIXES, BYTECODE_SUFFIXES
from itertools import chain
from pathlib import Path
from typing import Union


class ResourcesFolderError(Exception):
    pass


class ResourcesFolder:
    """
    A simple and universal utility to access data files in python packages.
    The only restriction is for the files to be in a separate folder that is not a python package. (for clarity)
    """

    __slots__ = ("path",)

    DEFAULT_DIR = "resources"

    def __init__(self, root_path: Union[Path, str], rel_dir_path: str = None):
        root_path = Path(root_path)
        self.path = root_path.joinpath(rel_dir_path or self.DEFAULT_DIR).resolve()
        if not self.path.is_dir():
            raise ResourcesFolderError(f"Resources folder is not a directory: {self.path}")
        try:
            self.path.relative_to(root_path)
        except ValueError:
            raise ResourcesFolderError(f"Resources folder must be a subdirectory under the root directory: {self.path}")
        if any((self.path / f"__init__{suffix}").exists() for suffix in chain(SOURCE_SUFFIXES, BYTECODE_SUFFIXES)):
            raise ResourcesFolderError(f"Resources folder must not be a python package to avoid ambiguity: {self.path}")

    def join_path(self, *parts: str) -> Path:
        return self.path.joinpath(*parts)

    def join(self, *parts: str) -> str:
        return str(self.join_path(*parts))
